/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 4.5 (TDR-T-convex), p. 99  
 *
 * this file: algo04_05.c
 *
 * For a demo how to use this algorithm see the file "demo_algo04_05.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"


/* for c=0 */
/*
#define T(aaa) log(aaa)
#define DT(aaa) (1./(aaa))
#define FT(aaa) exp(aaa)
#define FTI(aaa) log(aaa)
#define TI(aaa) exp(aaa)
*/

/* for c=1 */
#define T(aaa) (aaa)
#define DT(aaa) (1.)
#define FT(aaa) (0.5*(aaa)*(aaa))
#define FTI(aaa) sqrt(2.*(aaa))
#define TI(aaa) (aaa)


    static double *p=NULL,
      *tfp=NULL, *dtfp=NULL, *h=NULL, *b=NULL, *a=NULL, atotal, (*pdfh)(double);
    static double x;
    static int NN;

int setup04_05(  double (*pdf)(double), double (*dpdf)(double),  
                 int N, double constructionp[])
/*
 * Algorithm 4.5 (TDR-T-convex), p. 99  
 *
 * setup ... 1 = new setup; 0 = sample; -1 = free tables
 * pdf... function pointer to the density
 * dpdf... function pointer to the derivative of the density
 * N  ... number of construction points
 * constructionp[] ... array holding the construction points
 * constructionp[0] is the left, constructionp[N-1] the right border of the domain
 *
 */
{   
    int i;  

     if(p!=NULL){
       free(p);
       free(tfp);
       free(dtfp);
       free(h);
       free(b);
       free(a);
     }
     NN = N;
     pdfh =pdf;
     p = malloc((N+1)*sizeof(double));  
     tfp = malloc((N+1)*sizeof(double));  
     dtfp = malloc((N+1)*sizeof(double));  
     h = malloc((N+1)*sizeof(double));  
     b = malloc((N+1)*sizeof(double));  
     a = malloc((N+1)*sizeof(double));  
     for(i=1;i<=N;i++) p[i] = constructionp[i-1];
     for(i=1;i<=N;i++){
/*1*/  tfp[i] = T((*pdf)(p[i]));
       dtfp[i] = DT((*pdf)(p[i]))*(*dpdf)(p[i]);
       /* computes the derivative of the transformed density f-tilda'( p_i ) */   
     }
     for(i=1;i<N;i++){
/*2*/  h[i]=(tfp[i+1]-tfp[i])/(p[i+1]-p[i]); 
       /* printf("i %d h[i] %f p[i] %f tfp[i]%f\n",i,h[i],p[i],tfp[i]); */
/*3*/  b[i] = 0.5*(p[i+1]+p[i]);/*easier than to compute the intersections*/
     }
     b[0] = p[1];
     b[N] = p[N];
     atotal=0.;
     for(i=1;i<N;i++){
/*4*/  a[i] = (FT(tfp[i]+h[i]*(p[i+1]-p[i]))-FT(tfp[i]))/h[i];
/*5*/  atotal += a[i];
     }
     for(i=1;i<=N;i++){
       printf("i%d <%f>---%f area %f\n",i,p[i],b[i],a[i]);
       printf("tfp %f dtfp %f h %f\n",tfp[i],dtfp[i],h[i]);
     }
     printf("total area below the hat: %f\n",atotal);
     return 1;
}

double generator04_05()
/* is called after the setup to generate random variates */
{
   double u, v, asum, sx, hx;
    int j;  

/*6*/ while(1){
/*7*/   v = atotal*uniform();
/*8*/   j = 1;
	asum = a[1];
	while(v > asum){
	  j++;
	  asum += a[j];
	}
/*9*/   v = v - (asum-a[j]);
/*10*/  x = p[j]+(FTI(h[j]*v+FT(tfp[j]))-tfp[j])/h[j];
/*11*/  hx = TI(tfp[j]+h[j]*(x-p[j]));
/*12*/  if(x < b[j]) 
/*13*/    sx=TI(tfp[j]+dtfp[j]*(x-p[j]));
/*14*/  else 
/*15*/    sx=TI(tfp[j+1]+dtfp[j+1]*(x-p[j+1]));
/*16*/   u = uniform();
/*17*/   if(u*hx <= sx) 
/*18*/     return x;
/*19*/   if(u*hx <= (*pdfh)(x)) 
/*20*/     return x;
      }
}

int free04_05()
/* frees the allocated memory */ 
{
      free(p);
      p = NULL;
      free(tfp);
      tfp = NULL;
      free(dtfp);
      dtfp = NULL;
      free(h);
      h = NULL;
      free(b);
      b = NULL;
      free(a);
      a = NULL;
      return 1;
}


