/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 6.7 (Generalized Simple RoU, p. 147)
 *
 * this file: algo06_07.c
 *
 * For a demo how to use this algorithm see the file "demo_algo06_07.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"

#define POW(a,b) exp(log(a)*b)

static double ul, ur, vm, m, a, b, r, (*pdfh)(double);

int setup06_07( double (*pdf)(double), double Af, double mi, double Fm, double ri)
/*
 * Algorithm 6.7 (Generalized Simple RoU, p. 147)
 *
 * pdf ... function pointer to the quasi density
 * Af ... area below the quasi density
 * mi ... mode of the density
 * Fm ... CDF value at the mode m; if unknown -1. must be used
 * ri ... parameter of the method: we must have r > 1 and
 *       for the c of c-concavity we must have r >= -c/(1+c)
 */
{     double um, p;
      pdfh = pdf;
      m = mi;
      r = ri;
/*1*/ p = 1.-2.187/POW((r+5-1.28/r),0.9460);
/*2*/ b = (1.-r*POW(p,(r-1.))+(r-1.)*POW(p,r))/((POW(p,r)-1.)*(POW(p,r)-1.));
/*3*/ a = -(p-1.)/(POW(p,r)-1.)-p*b;
/*4*/ vm = POW((*pdfh)(m),1./(r+1.));
      um = Af/(r*vm);
/*5*/ if(Fm >= 0.){/* the cdf value at the mode is known */
/*6*/    ul = - Fm*um;
         ur = um+ul;
      }
/*7*/ else{
/*8*/    ul = -um;
         ur = um;
      }
      printf("Setup06_07:\n ul %f ur %f vm %f\n\n",ul,ur,vm);
      return 1;
} 

double generator06_07()
/* is called after the setup to generate random variates */
{ double u,v,w,x,z;

/*9*/ while(1){
/*10*/  w = log(a/(a+b))*uniform();
/*11*/  z = ul+(ur-ul)*uniform();
/*12*/  v = (exp(-w)-1.)*a/b;
/*13*/  u = -z/(a+b*v);
        v = v*vm;
/*14*/  x = u/POW(v,r)+m;
/*15*/  if(POW(v,(r+1.))<=(*pdfh)(x))
/*16*/    return x;
      }
}


