/*
 * Hormann, Leydold, Derflinger
 * Automatic Nonuniform Random Variate Generation
 * Springer, Heidelberg, New York, 2003
 * http://statistik.wu-wien.ac.at/arvag/monograph
 * 
 * Code for Algorithm 7.1 (Fast-Table-Inversion, p. 163)
 *
 * this file: algo07_01.c
 *
 * For a demo how to use this algorithm see the file "demo_algo07_01.c".
 *
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "uniform.h"


/* Insert Function into array... */
void insert(int location,int n,double m,double *array)
{  
int i;
for (i=n;i>location;i--)
    {
      array[i+1]=array[i];
    }
array[location+1]=m;

}


static int N,C,*gt=NULL;
static double *F=NULL,*b=NULL,*beta=NULL,*gam=NULL,*d=NULL;

int setup07_01( double (*cdf)(double), double (*pdf)(double),
                double eu, int nstart, int nmax, double bl, double br) 
/*
 * Algorithm 7.1 (Fast-Table-Inversion, p. 163)
 *
 * cdf ... function pointer to the cumulative distribution function
 * pdf ... function pointer to the density
 * eu ... maxmial allowed error in u-direction
 * nstart ... number of starting intervals
 * nmax ... maximal number of intervals possible
 * bl ... approximate left border of domain, pdf(bl)>0 is required
 * br ... approximate right border of domain, pdf(br)>0 is required
 */
{     int i,j,ns=nstart;   
      double *f;
      double error,k,fip1,Fip1;
      if(F!=NULL){
        free(F);
        free(b);
        free(beta);
        free(gam);
        free(d);
        free(gt);
      }
      F = malloc((nmax+1)*sizeof(double));  
      b = malloc((nmax+1)*sizeof(double));  
      beta = malloc((nmax+1)*sizeof(double));  
      gam = malloc((nmax+1)*sizeof(double));  
      d = malloc((nmax+1)*sizeof(double));  
      f = malloc((nmax+1)*sizeof(double));  
/*1*/ if((*pdf)(bl)<1.e-100){
        printf("setup07_01, ERROR, density-value for bl too small\n");
        exit(0);
      }
      if((*pdf)(br)<1.e-100){
        printf("setup07_01, ERROR, density-value for br too small\n");
	exit(0);
      }
/*2*/ for (i=0;i<=ns;i++){
/*3*/   b[i]=bl+i*(br-bl)/ns;
/*4*/	f[i]=(*pdf)(b[i]);
	F[i]=(*cdf)(b[i]);
      }
/*5*/ i=0; 
/*6*/ do{
/*7*/	beta[i]=(F[i+1]-F[i])/f[i];
	gam[i]=3*(b[i+1]-b[i])-(F[i+1]-F[i])*(2/f[i]+1/f[i+1]);
	d[i]=2*(b[i]-b[i+1])+(F[i+1]-F[i])*(1/f[i]+1/f[i+1]);
	error=fabs((*cdf)(b[i]+0.5*(beta[i]+0.5*(gam[i]+0.5*d[i])))-0.5*(F[i]+F[i+1]));   
/*8*/	if(error<eu) 
/*9*/     i = i+1;
/*10*/  else{
/*11*/	  k=0.5*(b[i]+b[i+1]);
	  insert(i,ns,k,b);
/*12*/	  fip1 = (*pdf)(b[i+1]);
	  Fip1 = (*cdf)(b[i+1]);
	  insert(i,ns,fip1,f);
	  insert(i,ns,Fip1,F);
	  ns++;
	  if(ns>nmax){

	    printf("setup07_01, ERROR, specified nmax too small"); 
	    printf("to reach requested error bound eu\n");
	    exit(0);
	  }
        }
/*13*/}while(!(b[i]==br));
/*14*/N=i;
      F = realloc(F,(N+1)*sizeof(double));  
      b = realloc(b,(N+1)*sizeof(double));  
      beta = realloc(beta,(N+1)*sizeof(double));  
      gam = realloc(gam,(N+1)*sizeof(double));  
      d = realloc(d,(N+1)*sizeof(double));  
      free(f);
      C=2*N;
      gt = malloc((C+1)*sizeof(int));  
/*15*/gt[0]=0;
      for(j=1,i=0;j<C;j++){
	while(F[i]<(j/(double)C)) i++;
	gt[j]=i;
      }
      printf("Setup07_01\n   For requested eu=%g the total number of intervals is N=%d\n\n",eu,N); 
      return 1;
}


double generator07_01()
/* is called after the setup to generate random variates */
{      double u,ut,X;
       int j,i;
/*16*/ u=uniform();
/*17*/ j=(int)(u*C);
       i=gt[j];
       while (u>F[i+1]) i++;
/*18*/ ut=(u-F[i])/(F[i+1]-F[i]);
/*19*/ X=b[i]+ut*(beta[i]+ut*(gam[i]+ut*d[i]));
/*20*/ return X;
}

int free07_01()
/* frees the allocated memory */
{
  free(F);
  F = NULL;
  free(b);
  b = NULL;
  free(beta);
  beta = NULL;
  free(gam);
  gam = NULL;
  free(d);
  d = NULL;
  free(gt);
  gt = NULL;
  return 1;
}

